package featureflag

// UserApplyPatchWithoutWorktree enables a worktree-free implementation of the
// UserApplyPatch RPC that applies patches via a temporary index instead of
// checking out a worktree and running git-am(1).
var UserApplyPatchWithoutWorktree = NewFeatureFlag(
	"user_apply_patch_without_worktree",
	"v19.2.0",
	"https://gitlab.com/gitlab-org/gitaly/-/work_items/7254",
	true,
)
